/* Subroutines used for expanding RISC-V builtins.
   Copyright (C) 2011-2020 Free Software Foundation, Inc.
   Contributed by Andrew Waterman (andrew@sifive.com).

This file is part of GCC.

GCC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3, or (at your option)
any later version.

GCC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING3.  If not see
<http://www.gnu.org/licenses/>.  */

#define IN_TARGET_CODE 1

#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "tm.h"
#include "rtl.h"
#include "tree.h"
#include "gimple-expr.h"
#include "memmodel.h"
#include "expmed.h"
#include "profile-count.h"
#include "optabs.h"
#include "recog.h"
#include "diagnostic-core.h"
#include "stor-layout.h"
#include "expr.h"
#include "langhooks.h"

/* Macros to create an enumeration identifier for a function prototype.  */
#define RISCV_FTYPE_NAME0(A) RISCV_##A##_FTYPE
#define RISCV_FTYPE_NAME1(A, B) RISCV_##A##_FTYPE_##B
#define RISCV_FTYPE_NAME2(A, B, C) RISCV_##A##_FTYPE_##B##_##C
#define RISCV_FTYPE_NAME3(A, B, C, D) RISCV_##A##_FTYPE_##B##_##C##_##D

/* Classifies the prototype of a built-in function.  */
enum riscv_function_type {
#define DEF_RISCV_FTYPE(NARGS, LIST) RISCV_FTYPE_NAME##NARGS LIST,
#include "config/riscv/riscv-ftypes.def"
#undef DEF_RISCV_FTYPE
  RISCV_MAX_FTYPE_MAX
};

/* Specifies how a built-in function should be converted into rtl.  */
enum riscv_builtin_type {
  /* The function corresponds directly to an .md pattern.  */
  RISCV_BUILTIN_DIRECT,

  /* Likewise, but with return type VOID.  */
  RISCV_BUILTIN_DIRECT_NO_TARGET
};

/* Declare an availability predicate for built-in functions.  */
#define AVAIL(NAME, COND)		\
 static unsigned int			\
 riscv_builtin_avail_##NAME (void)	\
 {					\
   return (COND);			\
 }

/* This structure describes a single built-in function.  */
struct riscv_builtin_description {
  /* The code of the main .md file instruction.  See riscv_builtin_type
     for more information.  */
  enum insn_code icode;

  /* The name of the built-in function.  */
  const char *name;

  /* Specifies how the function should be expanded.  */
  enum riscv_builtin_type builtin_type;

  /* The function's prototype.  */
  enum riscv_function_type prototype;

  /* Whether the function is available.  */
  unsigned int (*avail) (void);
};

AVAIL (bitmanip_zbb64, TARGET_ZBB && TARGET_64BIT)
AVAIL (bitmanip_zbc32, TARGET_ZBC && !TARGET_64BIT)
AVAIL (bitmanip_zbc64, TARGET_ZBC && TARGET_64BIT)
AVAIL (bitmanip_zbe32, TARGET_ZBE && !TARGET_64BIT)
AVAIL (bitmanip_zbe64, TARGET_ZBE && TARGET_64BIT)
AVAIL (bitmanip_zbp32, TARGET_ZBP && !TARGET_64BIT)
AVAIL (bitmanip_zbp64, TARGET_ZBP && TARGET_64BIT)
AVAIL (bitmanip_zbr32, TARGET_ZBR && !TARGET_64BIT)
AVAIL (bitmanip_zbr64, TARGET_ZBR && TARGET_64BIT)
AVAIL (bitmanip_zbt32, TARGET_ZBT && !TARGET_64BIT)
AVAIL (bitmanip_zbt64, TARGET_ZBT && TARGET_64BIT)
AVAIL (bitmanip_zbb_zbp32, (TARGET_ZBB || TARGET_ZBP) && !TARGET_64BIT)
AVAIL (bitmanip_zbb_zbp64, (TARGET_ZBB || TARGET_ZBP) && TARGET_64BIT)
AVAIL (bitmanip_zbp_zbe_zbf32, (TARGET_ZBP || TARGET_ZBE || TARGET_ZBF) && !TARGET_64BIT)
AVAIL (bitmanip_zbp_zbe_zbf64, (TARGET_ZBP || TARGET_ZBE || TARGET_ZBF) && TARGET_64BIT)
AVAIL (crypto_aes32e, TARGET_ZKNE && !TARGET_64BIT)
AVAIL (crypto_aes32d, TARGET_ZKND && !TARGET_64BIT)
AVAIL (crypto_sha, TARGET_ZKNH)
AVAIL (crypto_sha32, TARGET_ZKNH && !TARGET_64BIT)
AVAIL (hard_float, TARGET_HARD_FLOAT)
AVAIL (amethyst, TARGET_AMETHYST)

/* Construct a riscv_builtin_description from the given arguments.

   INSN is the name of the associated instruction pattern, without the
   leading CODE_FOR_riscv_.

   NAME is the name of the function itself, without the leading
   "__builtin_riscv_".

   BUILTIN_TYPE and FUNCTION_TYPE are riscv_builtin_description fields.

   AVAIL is the name of the availability predicate, without the leading
   riscv_builtin_avail_.  */
#define RISCV_BUILTIN(INSN, NAME, BUILTIN_TYPE,	FUNCTION_TYPE, AVAIL)	\
  { CODE_FOR_riscv_ ## INSN, "__builtin_riscv_" NAME,			\
    BUILTIN_TYPE, FUNCTION_TYPE, riscv_builtin_avail_ ## AVAIL }

/* Define __builtin_riscv_<INSN>, which is a RISCV_BUILTIN_DIRECT function
   mapped to instruction CODE_FOR_riscv_<INSN>,  FUNCTION_TYPE and AVAIL
   are as for RISCV_BUILTIN.  */
#define DIRECT_BUILTIN(INSN, FUNCTION_TYPE, AVAIL)			\
  RISCV_BUILTIN (INSN, #INSN, RISCV_BUILTIN_DIRECT, FUNCTION_TYPE, AVAIL)

/* Same as DIRECT_BUILTIN, but only define if we are on a 32bit target.  */
#define DIRECT_BUILTIN32(INSN, FUNCTION_TYPE, AVAIL)			\
  RISCV_BUILTIN (INSN ## si, #INSN, RISCV_BUILTIN_DIRECT, FUNCTION_TYPE, AVAIL ## 32)

/* Same as DIRECT_BUILTIN, but only define if we are on a 64bit target.  */
#define DIRECT_BUILTIN64(INSN, FUNCTION_TYPE, AVAIL)			\
  RISCV_BUILTIN (INSN ## di, #INSN, RISCV_BUILTIN_DIRECT, FUNCTION_TYPE, AVAIL ## 64)

/* Same as DIRECT_BUILTIN, but define single builtin in 32bit mode and two builtins in 64bit target.  */
#define DIRECT_BUILTINW(INSN, FUNCTION_TYPE32, FUNCTION_TYPE64, AVAIL)				\
  RISCV_BUILTIN (INSN ## si, #INSN, RISCV_BUILTIN_DIRECT, FUNCTION_TYPE32, AVAIL ## 32),	\
  RISCV_BUILTIN (INSN ## si, #INSN "w", RISCV_BUILTIN_DIRECT, FUNCTION_TYPE32, AVAIL ## 64),	\
  RISCV_BUILTIN (INSN ## di, #INSN, RISCV_BUILTIN_DIRECT, FUNCTION_TYPE64, AVAIL ## 64)

/* Same as DIRECT_BUILTIN, but define single builtin according to target word size.  */
#define DIRECT_BUILTINX(INSN, FUNCTION_TYPE32, FUNCTION_TYPE64, AVAIL)				\
  RISCV_BUILTIN (INSN ## si, #INSN, RISCV_BUILTIN_DIRECT, FUNCTION_TYPE32, AVAIL ## 32),	\
  RISCV_BUILTIN (INSN ## di, #INSN, RISCV_BUILTIN_DIRECT, FUNCTION_TYPE64, AVAIL ## 64)

/* Define __builtin_riscv_<INSN>, which is a RISCV_BUILTIN_DIRECT_NO_TARGET
   function mapped to instruction CODE_FOR_riscv_<INSN>,  FUNCTION_TYPE
   and AVAIL are as for RISCV_BUILTIN.  */
#define DIRECT_NO_TARGET_BUILTIN(INSN, FUNCTION_TYPE, AVAIL)		\
  RISCV_BUILTIN (INSN, #INSN, RISCV_BUILTIN_DIRECT_NO_TARGET,		\
		FUNCTION_TYPE, AVAIL)

/* Argument types.  */
#define RISCV_ATYPE_VOID void_type_node
#define RISCV_ATYPE_UQI unsigned_intQI_type_node
#define RISCV_ATYPE_QI intQI_type_node
#define RISCV_ATYPE_UHI unsigned_intHI_type_node
#define RISCV_ATYPE_HI intHI_type_node
#define RISCV_ATYPE_USI unsigned_intSI_type_node
#define RISCV_ATYPE_SI intSI_type_node
#define RISCV_ATYPE_UDI unsigned_intDI_type_node
#define RISCV_ATYPE_DI intDI_type_node

tree intV2HI_type_node;
tree intV2QI_type_node;
tree intV2SI_type_node;
tree intV4HI_type_node;
tree intV4QI_type_node;
tree intV8QI_type_node;

#define RISCV_ATYPE_V2HI intV2HI_type_node
#define RISCV_ATYPE_V2QI intV2QI_type_node
#define RISCV_ATYPE_V2SI intV2SI_type_node
#define RISCV_ATYPE_V4HI intV4HI_type_node
#define RISCV_ATYPE_V4QI intV4QI_type_node
#define RISCV_ATYPE_V8QI intV8QI_type_node

/* RISCV_FTYPE_ATYPESN takes N RISCV_FTYPES-like type codes and lists
   their associated RISCV_ATYPEs.  */
#define RISCV_FTYPE_ATYPES0(A) \
  RISCV_ATYPE_##A
#define RISCV_FTYPE_ATYPES1(A, B) \
  RISCV_ATYPE_##A, RISCV_ATYPE_##B

#define RISCV_FTYPE_ATYPES2(A, B, C) \
  RISCV_ATYPE_##A, RISCV_ATYPE_##B, RISCV_ATYPE_##C

#define RISCV_FTYPE_ATYPES3(A, B, C, D)	\
  RISCV_ATYPE_##A, RISCV_ATYPE_##B, RISCV_ATYPE_##C, RISCV_ATYPE_##D

static const struct riscv_builtin_description riscv_builtins[] = {
#include "config/riscv/amethyst-builtins.h"
  /* Bitmanip */
  DIRECT_BUILTIN (cpopw, RISCV_SI_FTYPE_SI, bitmanip_zbb64),
  DIRECT_BUILTIN (rolw, RISCV_SI_FTYPE_SI_SI, bitmanip_zbb_zbp64),
  DIRECT_BUILTINW (pack, RISCV_SI_FTYPE_HI_HI, RISCV_DI_FTYPE_SI_SI, bitmanip_zbp_zbe_zbf),
  DIRECT_BUILTINW (packu, RISCV_SI_FTYPE_SI_SI, RISCV_DI_FTYPE_DI_DI, bitmanip_zbp),
  DIRECT_BUILTINW (packh, RISCV_SI_FTYPE_QI_QI, RISCV_DI_FTYPE_QI_QI, bitmanip_zbp_zbe_zbf),
  DIRECT_BUILTINW (slo, RISCV_SI_FTYPE_SI_QI, RISCV_DI_FTYPE_DI_QI, bitmanip_zbb_zbp),
  DIRECT_BUILTINW (sro, RISCV_SI_FTYPE_SI_QI, RISCV_DI_FTYPE_DI_QI, bitmanip_zbb_zbp),
  DIRECT_BUILTINW (grev, RISCV_SI_FTYPE_SI_QI, RISCV_DI_FTYPE_DI_QI, bitmanip_zbp),
  DIRECT_BUILTINW (gorc, RISCV_SI_FTYPE_SI_QI, RISCV_DI_FTYPE_DI_QI, bitmanip_zbp),
  DIRECT_BUILTINW (shfl, RISCV_SI_FTYPE_SI_QI, RISCV_DI_FTYPE_DI_QI, bitmanip_zbp),
  DIRECT_BUILTINW (unshfl, RISCV_SI_FTYPE_SI_QI, RISCV_DI_FTYPE_DI_QI, bitmanip_zbp),
  DIRECT_BUILTINX (xpermn, RISCV_SI_FTYPE_SI_SI, RISCV_DI_FTYPE_DI_DI, bitmanip_zbp),
  DIRECT_BUILTINX (xpermb, RISCV_SI_FTYPE_SI_SI, RISCV_DI_FTYPE_DI_DI, bitmanip_zbp),
  DIRECT_BUILTINX (xpermh, RISCV_SI_FTYPE_SI_SI, RISCV_DI_FTYPE_DI_DI, bitmanip_zbp),
  DIRECT_BUILTIN64 (xpermw, RISCV_DI_FTYPE_DI_DI, bitmanip_zbp),
  DIRECT_BUILTINW (bmext, RISCV_SI_FTYPE_SI_SI, RISCV_DI_FTYPE_DI_DI, bitmanip_zbe),
  DIRECT_BUILTINW (bmdep, RISCV_SI_FTYPE_SI_SI, RISCV_DI_FTYPE_DI_DI, bitmanip_zbe),
  DIRECT_BUILTINX (clmul, RISCV_SI_FTYPE_SI_SI, RISCV_DI_FTYPE_DI_DI, bitmanip_zbc),
  DIRECT_BUILTINX (clmulh, RISCV_SI_FTYPE_SI_SI, RISCV_DI_FTYPE_DI_DI, bitmanip_zbc),
  DIRECT_BUILTINX (clmulr, RISCV_SI_FTYPE_SI_SI, RISCV_DI_FTYPE_DI_DI, bitmanip_zbc),
  DIRECT_BUILTINX (crc32b, RISCV_SI_FTYPE_SI, RISCV_DI_FTYPE_DI, bitmanip_zbr),
  DIRECT_BUILTINX (crc32h, RISCV_SI_FTYPE_SI, RISCV_DI_FTYPE_DI, bitmanip_zbr),
  DIRECT_BUILTINX (crc32w, RISCV_SI_FTYPE_SI, RISCV_DI_FTYPE_DI, bitmanip_zbr),
  DIRECT_BUILTIN64 (crc32d, RISCV_DI_FTYPE_DI, bitmanip_zbr),
  DIRECT_BUILTINX (crc32cb, RISCV_SI_FTYPE_SI, RISCV_DI_FTYPE_DI, bitmanip_zbr),
  DIRECT_BUILTINX (crc32ch, RISCV_SI_FTYPE_SI, RISCV_DI_FTYPE_DI, bitmanip_zbr),
  DIRECT_BUILTINX (crc32cw, RISCV_SI_FTYPE_SI, RISCV_DI_FTYPE_DI, bitmanip_zbr),
  DIRECT_BUILTIN64 (crc32cd, RISCV_DI_FTYPE_DI, bitmanip_zbr),
  DIRECT_BUILTINX (cmov, RISCV_SI_FTYPE_SI_SI_SI, RISCV_DI_FTYPE_DI_DI_DI, bitmanip_zbt),
  DIRECT_BUILTINW (fsl, RISCV_SI_FTYPE_SI_SI_SI, RISCV_DI_FTYPE_DI_DI_DI, bitmanip_zbt),
  DIRECT_BUILTINW (fsr, RISCV_SI_FTYPE_SI_SI_SI, RISCV_DI_FTYPE_DI_DI_DI, bitmanip_zbt),
  /* Crypto */
  DIRECT_BUILTIN (aes32esmi, RISCV_SI_FTYPE_SI_SI, crypto_aes32e),
  DIRECT_BUILTIN (aes32esi, RISCV_SI_FTYPE_SI_SI, crypto_aes32e),
  DIRECT_BUILTIN (aes32dsmi, RISCV_SI_FTYPE_SI_SI, crypto_aes32d),
  DIRECT_BUILTIN (aes32dsi, RISCV_SI_FTYPE_SI_SI, crypto_aes32d),
  DIRECT_BUILTIN (sha256sig0, RISCV_SI_FTYPE_SI, crypto_sha),
  DIRECT_BUILTIN (sha256sig1, RISCV_SI_FTYPE_SI, crypto_sha),
  DIRECT_BUILTIN (sha256sum0, RISCV_SI_FTYPE_SI, crypto_sha),
  DIRECT_BUILTIN (sha256sum1, RISCV_SI_FTYPE_SI, crypto_sha),
  DIRECT_BUILTIN (sha512sum0r, RISCV_SI_FTYPE_SI_SI, crypto_sha32),
  DIRECT_BUILTIN (sha512sum1r, RISCV_SI_FTYPE_SI_SI, crypto_sha32),
  DIRECT_BUILTIN (sha512sig0l, RISCV_SI_FTYPE_SI_SI, crypto_sha32),
  DIRECT_BUILTIN (sha512sig0h, RISCV_SI_FTYPE_SI_SI, crypto_sha32),
  DIRECT_BUILTIN (sha512sig1l, RISCV_SI_FTYPE_SI_SI, crypto_sha32),
  DIRECT_BUILTIN (sha512sig1h, RISCV_SI_FTYPE_SI_SI, crypto_sha32),
  /* Float */
  DIRECT_BUILTIN (frflags, RISCV_USI_FTYPE, hard_float),
  DIRECT_NO_TARGET_BUILTIN (fsflags, RISCV_VOID_FTYPE_USI, hard_float)
};

/* Index I is the function declaration for riscv_builtins[I], or null if the
   function isn't defined on this target.  */
static GTY(()) tree riscv_builtin_decls[ARRAY_SIZE (riscv_builtins)];

/* Get the index I of the function declaration for riscv_builtin_decls[I]
   using the instruction code or return null if not defined for the target.  */
static GTY(()) int riscv_builtin_decl_index[NUM_INSN_CODES];

#define GET_BUILTIN_DECL(CODE) \
  riscv_builtin_decls[riscv_builtin_decl_index[(CODE)]]

/* Return the function type associated with function prototype TYPE.  */

static tree
riscv_build_function_type (enum riscv_function_type type)
{
  static tree types[(int) RISCV_MAX_FTYPE_MAX];

  if (types[(int) type] == NULL_TREE)
    switch (type)
      {
#define DEF_RISCV_FTYPE(NUM, ARGS)					\
  case RISCV_FTYPE_NAME##NUM ARGS:					\
    types[(int) type]							\
      = build_function_type_list (RISCV_FTYPE_ATYPES##NUM ARGS,		\
				  NULL_TREE);				\
    break;
#include "config/riscv/riscv-ftypes.def"
#undef DEF_RISCV_FTYPE
      default:
	gcc_unreachable ();
      }

  return types[(int) type];
}

/* Implement TARGET_INIT_BUILTINS.  */

void
riscv_init_builtins (void)
{
  intV2HI_type_node = build_vector_type_for_mode (intHI_type_node, V2HImode);
  intV2QI_type_node = build_vector_type_for_mode (intQI_type_node, V2QImode);
  intV2SI_type_node = build_vector_type_for_mode (intSI_type_node, V2SImode);
  intV4HI_type_node = build_vector_type_for_mode (intHI_type_node, V4HImode);
  intV4QI_type_node = build_vector_type_for_mode (intQI_type_node, V4QImode);
  intV8QI_type_node = build_vector_type_for_mode (intQI_type_node, V8QImode);

  for (size_t i = 0; i < ARRAY_SIZE (riscv_builtins); i++)
    {
      const struct riscv_builtin_description *d = &riscv_builtins[i];
      if (d->avail ())
	{
	  tree type = riscv_build_function_type (d->prototype);
	  riscv_builtin_decls[i]
	    = add_builtin_function (d->name, type, i, BUILT_IN_MD, NULL, NULL);
	  riscv_builtin_decl_index[d->icode] = i;
	}
    }
}

/* Implement TARGET_BUILTIN_DECL.  */

tree
riscv_builtin_decl (unsigned int code, bool initialize_p ATTRIBUTE_UNUSED)
{
  if (code >= ARRAY_SIZE (riscv_builtins))
    return error_mark_node;
  return riscv_builtin_decls[code];
}

/* Take argument ARGNO from EXP's argument list and convert it into
   an expand operand.  Store the operand in *OP.  */

static void
riscv_prepare_builtin_arg (struct expand_operand *op, tree exp, unsigned argno)
{
  tree arg = CALL_EXPR_ARG (exp, argno);
  create_input_operand (op, expand_normal (arg), TYPE_MODE (TREE_TYPE (arg)));
}

/* Expand instruction ICODE as part of a built-in function sequence.
   Use the first NOPS elements of OPS as the instruction's operands.
   HAS_TARGET_P is true if operand 0 is a target; it is false if the
   instruction has no target.

   Return the target rtx if HAS_TARGET_P, otherwise return const0_rtx.  */

static rtx
riscv_expand_builtin_insn (enum insn_code icode, unsigned int n_ops,
			   struct expand_operand *ops, bool has_target_p)
{
  if (!maybe_expand_insn (icode, n_ops, ops))
    {
      error ("invalid argument to built-in function");
      return has_target_p ? gen_reg_rtx (ops[0].mode) : const0_rtx;
    }

  return has_target_p ? ops[0].value : const0_rtx;
}

/* Expand a RISCV_BUILTIN_DIRECT or RISCV_BUILTIN_DIRECT_NO_TARGET function;
   HAS_TARGET_P says which.  EXP is the CALL_EXPR that calls the function
   and ICODE is the code of the associated .md pattern.  TARGET, if nonnull,
   suggests a good place to put the result.  */

static rtx
riscv_expand_builtin_direct (enum insn_code icode, rtx target, tree exp,
			     bool has_target_p)
{
  struct expand_operand ops[MAX_RECOG_OPERANDS];

  /* Map any target to operand 0.  */
  int opno = 0;
  if (has_target_p)
    create_output_operand (&ops[opno++], target, TYPE_MODE (TREE_TYPE (exp)));

  /* Map the arguments to the other operands.  */
  gcc_assert (opno + call_expr_nargs (exp)
	      == insn_data[icode].n_generator_args);
  for (int argno = 0; argno < call_expr_nargs (exp); argno++)
    riscv_prepare_builtin_arg (&ops[opno++], exp, argno);

  return riscv_expand_builtin_insn (icode, opno, ops, has_target_p);
}

/* Implement TARGET_EXPAND_BUILTIN.  */

rtx
riscv_expand_builtin (tree exp, rtx target, rtx subtarget ATTRIBUTE_UNUSED,
		      machine_mode mode ATTRIBUTE_UNUSED,
		      int ignore ATTRIBUTE_UNUSED)
{
  tree fndecl = TREE_OPERAND (CALL_EXPR_FN (exp), 0);
  unsigned int fcode = DECL_MD_FUNCTION_CODE (fndecl);
  const struct riscv_builtin_description *d = &riscv_builtins[fcode];

  switch (d->builtin_type)
    {
    case RISCV_BUILTIN_DIRECT:
      return riscv_expand_builtin_direct (d->icode, target, exp, true);

    case RISCV_BUILTIN_DIRECT_NO_TARGET:
      return riscv_expand_builtin_direct (d->icode, target, exp, false);
    }

  gcc_unreachable ();
}

/* Implement TARGET_ATOMIC_ASSIGN_EXPAND_FENV.  */

void
riscv_atomic_assign_expand_fenv (tree *hold, tree *clear, tree *update)
{
  if (!TARGET_HARD_FLOAT)
    return;

  tree frflags = GET_BUILTIN_DECL (CODE_FOR_riscv_frflags);
  tree fsflags = GET_BUILTIN_DECL (CODE_FOR_riscv_fsflags);
  tree old_flags = create_tmp_var_raw (RISCV_ATYPE_USI);

  *hold = build4 (TARGET_EXPR, RISCV_ATYPE_USI, old_flags,
		  build_call_expr (frflags, 0), NULL_TREE, NULL_TREE);
  *clear = build_call_expr (fsflags, 1, old_flags);
  *update = NULL_TREE;
}
